import rich_click as click
from novara.utils import logger
import os
import toml
from novara.request import request, JSONDecodeError
import json


def generate_dockerfile():
    try:
        with open("novara.toml", "r") as f:
            # exploit_config = toml.load(f)
            toml_raw = f.read()
    except (OSError, FileNotFoundError):
        raise click.ClickException("novara.toml either not there or unaccessable")

    logger.info("read toml file")

    logger.info("generating Dockerfile")
    r = request.post(
        f"api/build/dockerfile/toml/", data=json.dumps(toml_raw)
    )
    if not r.ok:
        raise click.ClickException(
            f"failed generating dockerfile with message: {r.text}"
        )
        exit()

    try:
        dockerfile_str = r.json()
    except JSONDecodeError:
        raise click.ClickException(
            f"failed to decode response as json: {r.text[:20] if len(r.text) > 20 else r.text}"
        )

    try:
        with open("Dockerfile", "w") as docker_file:
            docker_file.write(dockerfile_str)
    except OSError:
        raise click.ClickException("Dockerfile is unaccessable")
        exit()

    logger.info("generated Dockerfile")
    logger.info("run 'novara run local|remote' to start a container or remote container")


@click.command()
@click.option("--pip", default=None, help="add a new pip packages")
@click.option("--apt", default=None, help="add a new apt packages")
@click.option("--apk", default=None, help="add a new apk packages")
@click.option("--snap", default=None, help="add a new snap packages")
@click.option("--dnf", default=None, help="add a new dnf packages")
@click.option("--directory", "-d", default="./", help="path to exploit")
def generate(pip, apt, apk, snap, dnf, directory):
    """(re)generate the Dockerfile and add dependencies if necessary"""

    directory = os.path.join(directory, "novara.toml")

    if any([pip, apt, apk, snap, dnf]):
        logger.info("adding packages to novara.toml file")

        try:
            with open(directory, "r") as f:
                exploit_config = toml.load(f)
        except (OSError, FileNotFoundError):
            raise click.ClickException(f"{directory} either not there or unaccessable")
            exit()

        docker_config = exploit_config.get("docker", None)
        if docker_config is None:
            raise click.ClickException("docker key not found in toml")
            exit()

        dependencies = docker_config.get("dependencies", None)
        if dependencies is None:
            raise click.ClickException("docker.dependencies key not found in toml")
            exit()

        pip_packages = dependencies.get("pip", None)
        apt_packages = dependencies.get("apt", None)
        apk_packages = dependencies.get("apk", None)
        dnf_packages = dependencies.get("dnf", None)
        snap_packages = dependencies.get("snap", None)

        if pip is not None:
            dependencies.update(
                {"pip": [pip] if pip_packages is None else pip_packages + [pip]}
            )

        if apt is not None:
            dependencies.update(
                {"apt": [apt] if apt_packages is None else apt_packages + [apt]}
            )

        if apk is not None:
            dependencies.update(
                {"apk": [apk] if apk_packages is None else apk_packages + [apk]}
            )

        if dnf is not None:
            dependencies.update(
                {"dnf": [dnf] if dnf_packages is None else dnf_packages + [dnf]}
            )

        if snap is not None:
            dependencies.update(
                {"snap": [snap] if snap_packages is None else snap_packages + [snap]}
            )

        try:
            with open(directory, "w") as f:
                f.write(
                    "# novara configuration file autogenerated by the novara cli, only edit if you know what you are doing\n\n"
                )
                toml.dump(exploit_config, f)
        except (OSError, FileNotFoundError):
            raise click.ClickException(f"{directory} either not there or unaccessable")
            exit()

        logger.info("updated novara.toml file")

    generate_dockerfile()



import rich_click as click
from urllib.parse import urljoin, urlparse
import os
import requests
from novara.request import AuthSession
from novara.config import Bootstrap_Config_Model, config, AuthConfig
from novara.utils import test_ssh_connection


@click.command()
@click.option(
    "--server_url",
    "-l",
    default=None,
    help="url to api-endpoint of your novara instance",
)
@click.option(
    "--username",
    "-u",
    default=None,
    help="username for basic auth. In the case of oauth this parameter does nothing.",
)
@click.option(
    "--password",
    "-p",
    default=None,
    help="password for basic auth. In the case of oauth this parameter does nothing.",
)
@click.option(
    "--author",
    "-a",
    default=None,
    help="to specify what author to use for the exploits",
)
def configure(server_url, username, password, author):
    """conect to novara backend & configure the cli"""

    # Priority: CLI argument > Environment variable > Prompt

    server_url = (
        server_url
        or os.environ.get("SERVER_URL")
        or click.prompt("Please enter the Novara server URL")
    )

    author = (
        author
        or os.environ.get("AUTHOR_NAME")
        or click.prompt("Please enter your author username")
    )

    r = requests.get(urljoin(server_url, '/api/config/auth_config/'))
    if not r.ok:
        raise click.ClickException(f"the remote responded with error:\n{r.text}")

    try:
        auth_config = AuthConfig(server_url=server_url, auth_server_url=urljoin(server_url,'/auth/'), **r.json())
        local_config = Bootstrap_Config_Model(author=author, server_url=server_url, auth_config=auth_config)
    except requests.JSONDecodeError:
        raise click.ClickException(f"unable to decode response as json:\n{r.text}")

    if local_config.auth_config.auth_type == "basic":
        parsed_server_url = urlparse(server_url)
        local_config.auth_config.username = (
            username
            or parsed_server_url.username
            or click.prompt("Please enter your username")
        )
        local_config.auth_config.password = (
            password
            or parsed_server_url.password
            or os.environ.get("PASSWORD")
            or click.prompt("Please enter your password")
        )

    session = AuthSession(local_config.auth_config)

    print('session configured')

    # -----------------------------------------------------------------

    r = session.get(urljoin(local_config.server_url, "/api/config/cli/"),
        params={'username':author}
    )
    if not r.ok:
        raise click.ClickException(f"the remote responded with error:\n{r.text}")

    # -----------------------------------------------------------------

    try:
        config.raw_write({**r.json(), **local_config.model_dump()})
    except requests.JSONDecodeError:
        raise click.ClickException(f"unable to decode response as json:\n{r.text}")
    
    test_ssh_connection()


from keycloak import KeycloakOpenID
from keycloak.exceptions import KeycloakPostError
from pydantic import BaseModel, Field
from typing import Optional, Union
from novara.config import AuthConfig, config
from requests.exceptions import *
from requests.auth import AuthBase, HTTPBasicAuth, HTTPDigestAuth
from urllib.parse import urljoin

import time
import webbrowser
import requests


class Token_Response(BaseModel):
    access_token:str
    refresh_token:str
    expires_in:int
    refresh_expires_in:int
    token_type:str
    session_state:str
    scope:str
    not_before_policy:int = Field(alias='not-before-policy')

class Device_Response(BaseModel):
    device_code:str
    user_code:str
    verification_uri:str
    verification_uri_complete:str
    expires_in:int
    interval:int

class UserInfoModel(BaseModel):
    id:Optional[str]
    username:Optional[str]
    email:Optional[str]

CLIENT_ID='novara-cli'

class KeycloakOIDCAuth(AuthBase):
    auth_config:AuthConfig
    client:KeycloakOpenID = None

    def __init__(self, auth_config:AuthConfig):
        self.auth_config = auth_config
        self.client = KeycloakOpenID(server_url=self.auth_config.auth_server_url, realm_name='oauth2-proxy', client_id='novara-cli')
        if self.auth_config.valide_until and self.auth_config.valide_until > time.time():
            return
        device_response = Device_Response.model_validate(self.client.device())

        webbrowser.open(device_response.verification_uri_complete)
        print(f'login at {device_response.verification_uri} with code {device_response.user_code}')

        for _ in range(device_response.expires_in // device_response.interval):
            try:
                raw_token = self.client.token(grant_type='urn:ietf:params:oauth:grant-type:device_code', device_code=device_response.device_code)
                break
            except KeycloakPostError as e:
                if b'authorization_pending' in e.error_message:
                    time.sleep(device_response.interval)
                    continue
                raise e

        raw_token['valide_until'] = time.time() + raw_token['refresh_expires_in']

        self.auth_config.update(raw_token)

    def refresh_auth_token(self):
        if time.time() < self.auth_config.valide_until:
            raw_tokens = self.client.refresh_token(self.auth_config.refresh_token)
            raw_tokens['valide_until'] = time.time() + raw_tokens['refresh_expires_in']
            self.auth_config.update(raw_tokens)
        
    def get_user_info(self) -> UserInfoModel:
        raw_user_info = self.client.userinfo(self.auth_config.access_token)
        return UserInfoModel(raw_user_info)

    def __call__(self, r: requests.PreparedRequest) -> requests.PreparedRequest:
        r.headers['Authorization'] = f'{self.auth_config.token_type} {self.auth_config.access_token}'
        return r

class AuthSession(requests.Session):
    auth:Union[KeycloakOIDCAuth, HTTPBasicAuth, HTTPDigestAuth, None] = None
    def __init__(self, auth_config:AuthConfig, **kwargs):
        super().__init__()
        self.auth_config = auth_config
        match auth_config.auth_type:
            case 'oauth':
                self.auth = KeycloakOIDCAuth(auth_config)
            case 'basic':
                self.auth = HTTPBasicAuth(auth_config.username, auth_config.password)
            case 'digest':
                self.auth = HTTPDigestAuth(auth_config.username, auth_config.password)
        
        self.on_update()

    def get_user_info(self):
        if hasattr(self.auth, 'get_user_info'):
            return self.auth.get_user_info()
        return UserInfoModel(name=self.auth_config.username, email=None)

    def refresh_token(self):
        if hasattr(self.auth, 'refresh_auth_token'):
            self.auth.refresh_auth_token()
            self.on_update()
        
    def on_update(self):
        pass

    def request(self,
                method,
                url,
                params=None,
                data=None,
                headers=None,
                cookies=None,
                files=None,
                auth=None,
                timeout=None,
                allow_redirects=True,
                proxies=None,
                hooks=None,
                stream=None,
                verify=None,
                cert=None,
                json=None):

        self.refresh_token()

        return super().request(method=method, 
                               url=url, 
                               params=params, 
                               data=data, 
                               headers=headers, 
                               cookies=cookies, 
                               files=files, 
                               auth=auth or self.auth, 
                               timeout=timeout,
                               allow_redirects=allow_redirects, 
                               proxies=proxies, 
                               hooks=hooks, 
                               stream=stream, 
                               verify=verify, 
                               cert=cert, 
                               json=json)

class LazyInitRequest(AuthSession):
    def __init__(self):
        self.is_initialized = False

    def on_update(self):
        config.save()

    def request(self, 
                method,
                url:str,
                params=None,
                data=None,
                headers=None,
                cookies=None,
                files=None,
                auth=None,
                timeout=None,
                allow_redirects=True,
                proxies=None,
                hooks=None,
                stream=None,
                verify=None,
                cert=None,
                json=None):
        
        if not url.startswith('http'):
            url = urljoin(config.auth_config.server_url, url)
        if not self.is_initialized:
            self.is_initialized = True
            super().__init__(config.auth_config)
        return super().request(method=method, 
                               url=url, 
                               params=params, 
                               data=data, 
                               headers=headers, 
                               cookies=cookies, 
                               files=files, 
                               auth=auth, 
                               timeout=timeout,
                               allow_redirects=allow_redirects, 
                               proxies=proxies, 
                               hooks=hooks, 
                               stream=stream, 
                               verify=verify, 
                               cert=cert, 
                               json=json)

request = LazyInitRequest()


from pydantic import BaseModel
import yaml
from typing import Union, Optional, Literal
import logging

from novara.constants import CONFIG_FILE, CONFIG_HOME

logger = logging.getLogger("rich")

class AuthConfig(BaseModel):
    auth_server_url:str
    server_url:Optional[str] = None
    auth_type:Literal['oauth', 'digest', 'basic']
    username:Optional[str] = None
    password:Optional[str] = None
    access_token:Optional[str] = None
    refresh_token:Optional[str] = None
    valide_until:Optional[float] = None
    token_type:Optional[str] = None
    expires_in:Optional[int] = None

    class Config:
        extra='allow'

    def update(self, updates:dict):
        for key, value in updates.items():
            setattr(self, key, value)

class Bootstrap_Config_Model(BaseModel):
    server_url:Optional[str]
    author:str
    auth_config:Optional[AuthConfig] = None

class Config_Model(BaseModel):
    server_url:Optional[str]
    author:str
    auth_type:str
    auth_config:Optional[AuthConfig] = None
    ssh_port:int
    ssh_user:str
    ssh_url:str
    ssh_privatekey:str


class ConfigManager(Config_Model):
    is_initialized: bool = False

    def __init__(self):
        object.__setattr__(self, 'is_initialized', False)       # avoid triggering loading logic

    def _load(self) -> dict:
        logger.info('loading new config...')
        try:
            with open(CONFIG_FILE, 'r') as config_file:
                return yaml.safe_load(config_file)
        except (FileNotFoundError, OSError):
            raise Exception()
            logger.error('config file not found or not accessible')
            logger.debug('did you run novara configure?')
            exit()
            
    def _initialize(self):
        super().__init__(**self._load())
        self.is_initialized = True
    
    def raw_write(self, config: dict):
        try:
            if not CONFIG_HOME.exists():
                logger.info(f"creating directory {CONFIG_HOME}")
                CONFIG_HOME.mkdir()
            config_directory = CONFIG_FILE.parent
            if not config_directory.exists():
                logger.info(f"creating directory {config_directory}")
                config_directory.mkdir()
            with open(CONFIG_FILE, 'w') as config_file:
                yaml.dump(config, config_file)
        except OSError:
            logger.error("Couldn't create the config file it's not writable")
            exit()

    def save(self):
        self.raw_write(self.raw_config)

    def __getattr__(self, name: str):
        if name in super().model_fields:
            self._initialize()

        return super().__getattribute__(name)

    @property
    def raw_config(self):
        """Access the config as a dict"""
        if not self.is_initialized:
            self._initialize()
        return self.model_dump()

    @raw_config.setter
    def raw_config(self, value: Union[dict, BaseModel]):
        if isinstance(value, BaseModel):
            value = value.model_dump()

        if self.is_initialized:
            value = {**self.model_dump(), **value}

        super().__init__(**value)

config = ConfigManager()